/*!\file Bamgx
 * \brief: use Bamg capabilities.
 */

#include "./Bamgx.h"

#undef __FUNCT__ 
#define __FUNCT__ "Bamgx"

#include "../shared/shared.h"
#include "../include/macros.h"
#include "../toolkits/toolkits.h"

/*Bamg: */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include <new>
#include <cassert>
#include <iomanip>
#include <fstream>
#include "Meshio.h"
#include "Mesh2.h"
#include "QuadTree.h"

using namespace bamg;
using namespace std;

int Bamgx(BamgMesh* bamgmesh,BamgGeom* bamggeom,BamgArgs* bamgargs){
	
	int noerr=1;

	/*output: */
	double* VerticesOut=NULL;
	double* TrianglesOut=NULL;
	int     NumVerticesOut;
	int     NumTrianglesOut;

	/*Bamg options*/
	int    iso;
	int    maxnbv;
	double hmin,hmax;
	double gradation;
	double cutoff;
	int    verbosity;
	int    splitcorners;

	/*Bamg variables: */
	int i,j,num;
	hinterpole=1; // by def interpolation a h 
	int fileout=0;
	int AbsError=0,nbjacoby=1,allquad=0;
	int Rescaling = 1;
	double costheta=2;
	Real8 cutoffradian=-1;
	double anisomax = 1e6;
	double err=0.01,errg=0.1,coef=1;
	int KeepBackVertices=1;
	double hminaniso=1e-100; 
	const double  boundmaxsubdiv = 10;
	double  maxsubdiv=boundmaxsubdiv;
	double omega=1.8;
	int NbSmooth=3;
	double *solMbb =0,*solMBB=0;
	int  *typesolsBB =0;
	Int4 nbsolbb=0,lsolbb=0;
	Int4 nbsolBB=0,lsolBB=0;
	int rbbeqMbb=0,rBBeqMBB=0;
	int ChoiseHessien = 0;
	double power=1;
	Triangles *Thr=0,*Thb=0;
	char *fmeshback=0,*fmeshout=0,*fmeshr=0,*fmetrix=0,*famfmt=0,*fmsh=0,
		 *fnopo=0,*fftq=0,*fam=0,*famdba=0,*rbb=0,*rBB=0,*wbb=0,*wBB=0,
		 *fMbb=0,*foM=0,*fMBB=0;
	char *datargv[128] ;
	int datargc=1;
	datargv[0]= datargv[1]=0;// for create a error if no parameter 

	/*Bamg options*/
	iso=bamgargs->iso;
	maxnbv=bamgargs->maxnbv;
	hmin=bamgargs->hmin;
	hmax=bamgargs->hmax;
	gradation=bamgargs->gradation;
	cutoff=bamgargs->cutoff;
	verbosity=bamgargs->verbose;
	splitcorners=bamgargs->splitcorners;

	// some verification
	if ( maxsubdiv > boundmaxsubdiv || maxsubdiv <= 1.0)
	{
		cerr << " -maxsubdiv " << maxsubdiv << " is not in ]1,"<< boundmaxsubdiv << "]" << endl;
		exit(3);
	}
	if (iso) anisomax=1;
	if (!(fmetrix||fMbb))
		NbSmooth=0; // no metric -> no smoothing 
	if( ! rbb) // to set the rbb filename by default 
		rbb=fMbb;
	if( ! rBB) // to set the rbb filename by default 
		rBB=fMBB;

	if (fMbb && rbb) 
		rbbeqMbb = !strcmp(rbb,fMbb);
	if (fMBB && rBB) 
		rBBeqMBB = !strcmp(rBB,fMBB);

	if(bamgmesh->NumTriangles==0){
		/*Mesh generation {{{1*/

		if (verbosity>0) printf("Construction of a mesh from a given geometry\n");
		if (verbosity>1) printf("   Processing geometry...\n");
		Geometry Gh(bamggeom,bamgargs);
		hmin = Max(hmin,Gh.MinimalHmin());
		hmax = Min(hmax,Gh.MaximalHmax());

		//build metric if not given in input
		if (verbosity>1) printf("   Generating Metric...\n");
		for(i=0;i<Gh.nbv;i++)
		  {
			MetricAnIso M=Gh[i];
			MatVVP2x2 Vp(M/coef);
			Vp.Maxh(hmin);
			Vp.Minh(hmax);
			Gh.vertices[i].m = Vp;
		  }

		//generate mesh
		if (verbosity>1) printf("   Generating Mesh...\n");
		Triangles Th(maxnbv,Gh);
		if(splitcorners) Th.SplitInternalEdgeWithBorderVertices();
		Th.ReNumberingTheTriangleBySubDomain();
		//if(NbSmooth>0) Th.SmoothingVertex(NbSmooth,omega);
		Th.SmoothingVertex(3,omega);

		//Build output
		if (verbosity>1) printf("   Processing Output...\n");
		NumVerticesOut=Th.nbv;
		VerticesOut=(double*)xmalloc(3*NumVerticesOut*sizeof(double));
		for (i=0;i<NumVerticesOut;i++){
			VerticesOut[i*3+0]=Th.vertices[i].r.x;
			VerticesOut[i*3+1]=Th.vertices[i].r.y;
			VerticesOut[i*3+2]=Th.vertices[i].ReferenceNumber;
		}

		NumTrianglesOut=Th.nbt-Th.NbOutT; //number of triangles - number of external triangles
		TrianglesOut=(double*)xmalloc(4*NumTrianglesOut*sizeof(double));
		num=0;
		for(i=0;i<Th.nbt;i++){ 
			Triangle &t=Th.triangles[i];
			if (t.link){
				//write the element only if it is part of the mesh (no boundary element)
				TrianglesOut[4*num+0]=Th.Number(t[0])+1;
				TrianglesOut[4*num+1]=Th.Number(t[1])+1;
				TrianglesOut[4*num+2]=Th.Number(t[2])+1;
				TrianglesOut[4*num+3]=t.color;
				num=num+1;
			}
		}
		/*}}}*/
	}

	else{
		/*Anisotropic mesh adaptation {{{1*/

		// read background mesh 
		if (verbosity>0) printf("Anisotropic mesh adaptation\n");
		if (verbosity>1) printf("   Processing initial mesh and geometry...\n");
		Triangles BTh(bamgmesh,bamgargs); 
		hmin = Max(hmin,BTh.MinimalHmin());
		hmax = Min(hmax,BTh.MaximalHmax());

		//?????TEST
		BTh.MakeQuadTree();

		//build metric if not given in input
		if (verbosity>1) printf("   Processing Metric...\n");
		if (bamgargs->metric){
			BTh.ReadMetric(bamgargs,hmin,hmax,coef);
		}
		else { // init with hmax 
			Metric Mhmax(hmax);
			for (Int4 iv=0;iv<BTh.nbv;iv++)
			 BTh[iv].m = Mhmax;
		}

		BTh.IntersectGeomMetric(errg,iso);
		if(gradation) BTh.SmoothMetric(gradation);
		BTh.MaxSubDivision(maxsubdiv);
		BTh.BoundAnisotropy(anisomax,hminaniso);

		//Build new mesh Thr
		Thr=&BTh,Thb=0;
		Triangles & Th( *(0 ?  new Triangles(*Thr,&Thr->Gh,Thb,maxnbv) :  new Triangles(maxnbv,BTh,KeepBackVertices)));
		if (Thr != &BTh) delete Thr;

		if(costheta<=1.0){
			Th.MakeQuadrangles(costheta);
		}
		if (allquad){
			Th.SplitElement(allquad==2);
		}
		if(splitcorners){
			Th.SplitInternalEdgeWithBorderVertices();
		}
		Th.ReNumberingTheTriangleBySubDomain();

		if(NbSmooth>0) Th.SmoothingVertex(NbSmooth,omega);

		Th.BTh.ReMakeTriangleContainingTheVertex();

		//info
		if(verbosity>0) {
			if (Th.nbt-Th.NbOutT-Th.NbOfQuad*2){
				printf("   new number of triangles = %i\n",(Th.nbt-Th.NbOutT-Th.NbOfQuad*2));
			}
			if (Th.NbOfQuad ){
				printf("   new number of quads = %i\n",Th.NbOfQuad);
			}
		}

		//Build output
		NumVerticesOut=Th.nbv;
		VerticesOut=(double*)xmalloc(3*NumVerticesOut*sizeof(double));
		for (i=0;i<NumVerticesOut;i++){
			VerticesOut[i*3+0]=Th.vertices[i].r.x;
			VerticesOut[i*3+1]=Th.vertices[i].r.y;
			VerticesOut[i*3+2]=Th.vertices[i].ReferenceNumber;
		}

		NumTrianglesOut=Th.nbt-Th.NbOutT; //number of triangles - number of external triangles
		TrianglesOut=(double*)xmalloc(4*NumTrianglesOut*sizeof(double));
		num=0;
		for(i=0;i<Th.nbt;i++){ 
			Triangle &t=Th.triangles[i];
			if (t.link){
				//write the element only if it is part of the mesh (no boundary element)
				TrianglesOut[4*num+0]=Th.Number(t[0])+1;
				TrianglesOut[4*num+1]=Th.Number(t[1])+1;
				TrianglesOut[4*num+2]=Th.Number(t[2])+1;
				TrianglesOut[4*num+3]=t.color;
				num=num+1;
			}
		}

		/*clean up*/
		//delete &Th;
		/*}}}*/
	}

	/*Assign output pointers*/
	bamgmesh->NumTriangles=NumTrianglesOut;
	xfree((void**)&bamgmesh->Triangles);
	bamgmesh->Triangles=TrianglesOut;

	bamgmesh->NumVertices=NumVerticesOut;
	xfree((void**)&bamgmesh->Vertices);
	bamgmesh->Vertices=VerticesOut;

	/*No error return*/
	if (verbosity>1) printf("   Exiting Bamg.\n");
	return noerr;

}
